'''
Copyright (C) 2020  Davis Remmel
Copyright 2021 Robert Schroll

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from .generic import GenericPen
from .textures import PENCIL_TEXTURES

class MechanicalPencilPen(GenericPen):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.vector = kwargs.get('vector', False)

    def set_segment_properties(self, canvas, segment, nextsegment):
        # Set the width
        canvas.setLineWidth(segment.width / 1.5)

        # Set the brush/pattern
        if self.vector:
            stroke_color = [1 - (1 - c) * segment.pressure for c in self.color]
            canvas.setStrokeColor(stroke_color)
        else:
            assert False
            brush.setColor(self.color())
            texture = PENCIL_TEXTURES.get_linear(0.00)
            pressure_textures = [
                PENCIL_TEXTURES.get_linear(0.10),
                PENCIL_TEXTURES.get_linear(0.15),
                PENCIL_TEXTURES.get_linear(0.20),
                PENCIL_TEXTURES.get_linear(0.25),
                PENCIL_TEXTURES.get_linear(0.30),
                PENCIL_TEXTURES.get_linear(0.40),
                PENCIL_TEXTURES.get_linear(0.50),
                PENCIL_TEXTURES.get_linear(0.60),
                PENCIL_TEXTURES.get_linear(0.70),
                PENCIL_TEXTURES.get_linear(0.80),
                PENCIL_TEXTURES.get_linear(0.90)
            ]
            for n, tex in enumerate(pressure_textures):
                threshold = n / len(pressure_textures)
                if segment.pressure >= threshold:
                    texture = tex
            brush.setTextureImage(texture)
            self.setBrush(brush)



'''
Copyright (C) 2020  Davis Remmel
Copyright 2021 Robert Schroll

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import logging
import tempfile
from pathlib import Path
import json
import re

from pdfrw import PdfReader, PdfWriter, PageMerge, PdfDict, PdfArray, PdfName, \
    IndirectPdfDict, uncompress, compress

from reportlab.pdfgen import canvas

from . import document, sources
from .constants import PDFHEIGHT, PDFWIDTH, PTPERPX, SPOOL_MAX


log = logging.getLogger(__name__)

def render(source, *, progress_cb=lambda x: None):
    # Exports the self as a PDF document to disk

    # progress_cb will be called with a progress percentage between 0 and
    # 100.  This percentage calculation is split 50% for the rendering
    # of the lines and 50% merging with the base PDF file.  This callback
    # also provides an opportunity to abort the process. If the callback
    # raises an error, this function will take steps to abort gracefullly
    # and pass the error upwards.

    vector=True  # TODO: Different rendering styles
    source = sources.get_source(source)

    # If this is using a base PDF, the percentage is calculated
    # differently.
    uses_base_pdf = source.exists('{ID}.pdf')

    # Document metadata should already be loaded (from device)
    # ...

    # Generate page information
    # If a PDF file was uploaded, but never opened, there may not be
    # a .content file. So, just load a barebones one with a 'pages'
    # key of zero length, so it doesn't break the rest of the
    # process.
    pages = []
    if source.exists('{ID}.content'):
        with source.open('{ID}.content', 'r') as f:
            pages = json.load(f).get('pages', [])

    # Render each page as a pdf
    tmpfh = tempfile.TemporaryFile()
    pdf_canvas = canvas.Canvas(tmpfh, (PDFWIDTH, PDFHEIGHT))
    # TODO: check pageCompression

    # Don't load all the pages into memory, because large notebooks
    # about 500 pages could use up to 3 GB of RAM. Create them by
    # iteration so they get released by garbage collector.
    changed_pages = []
    annotations = []
    for i in range(0, len(pages)):
        page = document.DocumentPage(source, pages[i], i)
        if source.exists(page.rmpath):
            changed_pages.append(i)
        page.render_to_painter(pdf_canvas, vector)
        annotations.append(page.get_grouped_annotations())
        progress_cb((i + 1) / len(pages) * 50)
    pdf_canvas.save()
    tmpfh.seek(0)

    # This new PDF represents just the notebook. If there was a
    # parent PDF, merge it now.
    if uses_base_pdf and not changed_pages:
        # Since there is no stroke data, just return the PDF data
        progress_cb(100)

        log.info('exported pdf')
        return source.open('{ID}.pdf', 'rb')

    # PDF exists, stroke data exists, so mix them together.
    if uses_base_pdf:
        rmpdfr = PdfReader(tmpfh)
        basepdfr = PdfReader(source.open('{ID}.pdf', 'rb'))
    else:
        basepdfr = PdfReader(tmpfh)
        # Alias, which is used for annotations and layers.
        rmpdfr = basepdfr

    # If making a 'layered' PDF (with optional content groups,
    # OCGs), associate the annoatations with the layer.

    # This property list is put into the rmpdfr document, which
    # will not have any existing properties.
    ocgprop = IndirectPdfDict(
        OCGs=PdfArray(),
        D=PdfDict(Order=PdfArray()))

    for i in range(0, len(basepdfr.pages)):
        basepage = basepdfr.pages[i]
        rmpage = rmpdfr.pages[i]

        # Apply OCGs
        apply_ocg = False #TODO configurable? bool(int(QSettings().value(
            #'pane/notebooks/export_pdf_ocg')))
        if apply_ocg:
            ocgorderinner = do_apply_ocg(basepage, rmpage, i, uses_base_pdf, ocgprop, annotations)
        else:
            ocgorderinner = None

        # Apply annotations to the rmpage. This must come after
        # applying OCGs, because the annotation may belong to
        # one of those groups.
        apply_annotations(rmpage, annotations[i], ocgorderinner)

        # If this is a normal notebook with highlighting,
        # just add the annotations and forget about the rest,
        # which are page geometry transformations.
        if uses_base_pdf:
            merge_pages(basepage, rmpage, i in changed_pages)

        progress_cb(((i + 1) / rmpdfr.numPages * 50) + 50)

    # Apply the OCG order. The basepdf may have already had OCGs
    # and so we must not overwrite them. NOTE: there are other
    # properties that ought to be carried over, but this is the
    # minimum required.
    if apply_ocg:
        if '/OCProperties' in basepdfr.Root:
            basepdfr.Root.OCProperties.OCGs += ocgprop.OCGs
            basepdfr.Root.OCProperties.D.Order += ocgprop.D.Order
        else:
            basepdfr.Root.OCProperties = ocgprop

    pdfw = PdfWriter()
    stream = tempfile.SpooledTemporaryFile(SPOOL_MAX)
    pdfw.write(stream, basepdfr)
    stream.seek(0)

    log.info('exported pdf')
    return stream


def do_apply_ocg(basepage, rmpage, i, uses_base_pdf, ocgprop, annotations):
    ocgpage = IndirectPdfDict(
        Type=PdfName('OCG'),
        Name='Page ' + str(i+1))
    ocgprop.OCGs.append(ocgpage)

    # The Order dict is a Page, followed by Inner
    ocgorderinner = PdfArray()


    # Add Template OCG layer
    # If this uses a basepdf, the template is located
    # elsewhere.


    # If using a basepdf, assign its stream as a
    # 'Background' layer under this page. When the page
    # primary OCG is disabled, the background will
    # remain, making it easy to disable all annotations.
    if uses_base_pdf:
        ocgorigdoc = IndirectPdfDict(
            Type=PdfName('OCG'),
            Name='Background')
        ocgprop.OCGs.append(ocgorigdoc)
        ocgorderinner.append(ocgorigdoc)

        uncompress.uncompress([basepage.Contents])
        stream = basepage.Contents.stream
        stream = '/OC /ocgorigdoc BDC\n' \
            + stream \
            + 'EMC\n'
        basepage.Contents.stream = stream
        compress.compress([basepage.Contents])

        if '/Properties' in basepage.Resources:
            props = basepage.Resources.Properties
        else:
            props = PdfDict()
        props.ocgorigdoc = ocgorigdoc
        basepage.Resources.Properties = props


    # If not using a basepdf, assign the rmpage's stream
    # as a 'Template' layer under this page. It will be
    # affected by disabling the primary Page OCG (which
    # by itself is kind of useless for exported
    # notebooks).

    # Regardless of using a basepdf or not, put the
    # rmpage layers into their own OCGs.

    # If the template has an XObject, we want to skip
    # the first one. This happens when the template
    # contains a PNG. Question--what happens when the
    # template contains more than one PNG? How do we
    # detect all of those?

    template_xobj_keys = []
    vector_layers = []
    uncompress.uncompress([rmpage.Contents])
    if uses_base_pdf:
        # The entire thing is the page ocg
        stream = '/OC /ocgpage BDC\n'
        stream += rmpage.Contents.stream
        stream += 'EMC\n'
        rmpage.Contents.stream = stream
    else:
        stream = rmpage.Contents.stream
        # Mark the template ocg separate from page ocg
        template_endpos = 0
        page_inatpos = 0
        findkey = '1 w 2 J 2 j []0  d\nq\n'
        # Finds only the first instance, which should be
        # for the template.
        findloc = stream.find(findkey)
        if findloc < 0:
            # May be a vector, which we stick a marker
            # in for.
            # ?? Why is this a half-point off ??
            findkey = '799.500000 85 l\n'
            m = re.search(
                findkey,
                rmpage.Contents.stream)
            if m:
                findloc = m.start()
        if findloc > 0:
            template_endpos = findloc + len(findkey)
            # Add vector template OCG
            stream = '/OC /ocgtemplate BDC\n'
            stream += rmpage.Contents.stream[:template_endpos]
            stream += 'EMC\n'
            page_inatpos = len(stream)
            stream += rmpage.Contents.stream[template_endpos:]
            # Save stream
            rmpage.Contents.stream = stream

        # Add template ocg
        ocgtemplate = IndirectPdfDict(
            Type=PdfName('OCG'),
            Name='Template')
        ocgprop.OCGs.append(ocgtemplate)
        ocgorderinner.append(ocgtemplate)

        # If a template (which is SVG) has embedded PNG
        # images, those appear as XObjects. This will
        # mess up the layer order, so we will ignore
        # them later.
        template_xobj_keys = \
            re.findall(r'(\/Im[0-9]+)\s',
                        stream[:template_endpos])

        # Page ocg
        stream = rmpage.Contents.stream[:page_inatpos]
        stream += '/OC /ocgpage BDC\n'
        stream += rmpage.Contents.stream[page_inatpos:]
        stream += 'EMC\n'
        # Save stream
        rmpage.Contents.stream = stream

    # Find all other vector layers using the magic
    # point (DocumentPageLayer.render_to_painter()).
    # ?? Why is this a half-point off ??
    while True:
        m = re.search(
            '420.500000 69 m\n',
            rmpage.Contents.stream)
        if not m:
            break
        stream = ''
        layerid = 'ocglayer{}'.format(
            len(vector_layers) + 1)
        stream = rmpage.Contents.stream[:m.start()]
        if len(vector_layers):
            # close previous layer
            stream += 'EMC\n'
        stream += '/OC /{} BDC\n'.format(layerid)
        stream += rmpage.Contents.stream[m.end():]
        vector_layers.append(layerid)
        rmpage.Contents.stream = stream
    # If we added vector layers, have to end the
    # first one.
    if len(vector_layers):
        stream = rmpage.Contents.stream + 'EMC\n'
        rmpage.Contents.stream = stream

    # Done--recompress the stream.
    compress.compress([rmpage.Contents])

    # There shouldn't be any Properties there since we
    # generated the rmpage ourselves, so don't bother
    # checking.
    rmpage.Resources.Properties = PdfDict(
        ocgpage=ocgpage)
    if not uses_base_pdf:
        rmpage.Resources.Properties.ocgtemplate = ocgtemplate

    # Add individual OCG layers (Bitmap)
    was_vector = True
    for n, key in enumerate(rmpage.Resources.XObject):
        if str(key) in template_xobj_keys:
            continue
        was_vector = False
        l = n - len(template_xobj_keys)
        # This would indicate a bug in the handling of a
        # notebook.
        try:
            layer = annotations[i][l]
        except:
            log.error('could not associate XObject with layer: (i, l) ({}, {})'.format(i, l))
            log.error(str(annotations))
            log.error('document: {} ()').format(
                'uuid',
                'self.visible_name')
            continue
        layername = layer[0]
        ocg = IndirectPdfDict(
            Type=PdfName('OCG'),
            Name=layername)
        ocgprop.OCGs.append(ocg)
        ocgorderinner.append(ocg)
        rmpage.Resources.XObject[key].OC = ocg

    # Add individual OCG layers (Vector)
    if was_vector:
        for l, layerid in enumerate(vector_layers):
            # This would indicate a bug in the handling of a
            # notebook.
            try:
                layer = annotations[i][l]
            except:
                log.error('could not associate layerid with layer: (i, l, layerid) ({}, {}, {})'.format(i, l, layerid))
                log.error('document: {} ()').format(
                    'uuid',
                    'self.visible_name')
                log.error(str(annotations))
                continue
            layername = layer[0]
            ocg = IndirectPdfDict(
                Type=PdfName('OCG'),
                Name=layername)
            ocgprop.OCGs.append(ocg)
            ocgorderinner.append(ocg)
            rmpage.Resources.Properties[PdfName(layerid)] = \
                ocg

    # Add order of OCGs to primary document
    ocgprop.D.Order.append(ocgpage)
    ocgprop.D.Order.append(ocgorderinner)

    return ocgorderinner


def apply_annotations(rmpage, page_annot, ocgorderinner):
    for k, layer_a in enumerate(page_annot):
        layerannots = layer_a[1]
        for a in layerannots:
            # PDF origin is in bottom-left, so invert all
            # y-coordinates.
            author = 'RCU' #self.model.device_info['rcuname']
            pdf_a = PdfDict(Type=PdfName('Annot'),
                            Rect=PdfArray([
                                (a[1] * PTPERPX),
                                PDFHEIGHT - (a[2] * PTPERPX),
                                (a[3] * PTPERPX),
                                PDFHEIGHT - (a[4] * PTPERPX)]),
                            T=author,
                            ANN='pdfmark',
                            Subtype=PdfName(a[0]),
                            P=rmpage)
            # Set to indirect because it makes a cleaner PDF
            # output.
            pdf_a.indirect = True
            if ocgorderinner:
                pdf_a.OC = ocgorderinner[k]
            if not '/Annots' in rmpage:
                rmpage.Annots = PdfArray()
            rmpage.Annots.append(pdf_a)


def merge_pages(basepage, rmpage, changed_page):
    # The general appraoch is to keep the base PDF. So, all
    # operations must be made upon the basepage. PyPDF2 will
    # keep all those pages' metadata and annotations,
    # including the paper size. However, a few things must
    # also occur.

    # The basepage must be reisized to the ratio of the rM
    # page so that no brush strokes get cut.

    # The new (rM) page must be resized to the dimensions of
    # the basepage. The PDF standard allows different page
    # sizes in one document, so each page must be measured.

    # ...

    # There is a bug here that can be seen with the NH file
    # It is possible (why?) for a page not to have a
    # MediaBox, so one must be taken from the parent. The
    # rM adds a bit to the width AND the height on this
    # file.
    bpage_box = basepage.MediaBox
    if not bpage_box:
        # Should probably check if the parent has a mediabox
        bpage_box = basepage.Parent.MediaBox
    bpage_w = float(bpage_box[2]) - float(bpage_box[0])
    bpage_h = float(bpage_box[3]) - float(bpage_box[1])
    # Round because floating point makes it prissy
    bpage_ratio = round(bpage_w / bpage_h * 10000) / 10000
    landscape_bpage = False
    if bpage_w > bpage_h:
        landscape_bpage = True

    # If the base PDF page was really wide, the rM rotates
    # it -90deg (CCW) on the screen, but doesn't actually
    # rotate it in the PDF. Also, if a notebook is in
    # landscape format, it remains in portrait mode during
    # the Web UI export. So, we must actually rotate the rM
    # page 90deg (CW) to fit on these wide pages.

    rpage_box = rmpage.MediaBox
    rpage_w = float(rpage_box[2]) - float(rpage_box[0])
    rpage_h = float(rpage_box[3]) - float(rpage_box[1])
    rpage_ratio = rpage_w / rpage_h
    if landscape_bpage:
        rmpage.Rotate = 90
        rpage_ratio = rpage_h / rpage_w

        # Annotations must be rotated because this rotation
        # statement won't hit until the page merge, and
        # pdfrw is unaware of annotations.
        if '/Annots' in rmpage:
            for a, annot in enumerate(rmpage.Annots):
                rect = annot.Rect
                rmpage.Annots[a].Rect = PdfArray([
                    rect[1],
                    PDFWIDTH - rect[0],
                    rect[3],
                    PDFWIDTH - rect[2]])


    # Resize the base page to the notebook page ratio by
    # adjusting the trimBox. If the basepage was landscape,
    # the trimbox must expand laterally, because the rM
    # rotates the page on-screen into portrait. If the
    # basepage was already portrait, it must expand
    # laterally.

    adjust = 0
    if bpage_ratio <= rpage_ratio:
        # Basepage is taller, so need to expand the width.
        # The basepage should be pushed to the right, which
        # is also the top of the rM in portrait mode. A
        # push to the right is really just decreasing the
        # left side.
        new_width = rpage_ratio * bpage_h
        if landscape_bpage:
            adjust = float(bpage_box[2]) - new_width
            bpage_box[0] = adjust
        else:
            # Portrait documents get pushed to the left, so
            # expand the right side.
            adjust = float(bpage_box[0])
            bpage_box[2] = new_width + float(bpage_box[0])
    elif bpage_ratio > rpage_ratio:
        # Basepage is fatter, so need to expand the height.
        # The basepage should be pushed to the top, which is
        # also the top of the rM in portrait mode. A push to
        # the top is really decreasing the bottom side.
        new_height = (1 / rpage_ratio) * bpage_w
        adjust = float(bpage_box[3]) - new_height
        bpage_box[1] = adjust

    # If this wasn't a changed page, don't bother with the
    # following.
    if not changed_page:
        return

    # Scale and (if necesssary) rotate the notebook page
    # and overlay it to the basepage. Might have to push
    # it a bit, depending on the direction.
    #basepage.Rotate = -90
    np = PageMerge(basepage).add(rmpage)

    annot_adjust = [0, 0]

    if bpage_ratio <= rpage_ratio:
        scale = bpage_h / np[1].h
        np[1].scale(scale)
        np[1].x = adjust
        annot_adjust[0] = adjust
    elif bpage_ratio > rpage_ratio:
        scale = bpage_w / np[1].w
        np[1].scale(scale)
        np[1].y = adjust
        annot_adjust[1] = adjust

    if '/Annots' in rmpage:
        for a, annot in enumerate(rmpage.Annots):
            rect = annot.Rect
            newrect = PdfArray([
                rect[0] * scale + annot_adjust[0],
                rect[1] * scale + annot_adjust[1],
                rect[2] * scale + annot_adjust[0],
                rect[3] * scale + annot_adjust[1]])
            rmpage.Annots[a].Rect = newrect

    # Gives the basepage the rmpage as a new object
    np.render()

    # Annots aren't carried over--pdfrw isn't aware.
    if '/Annots' in rmpage:
        if not '/Annots' in basepage:
            basepage.Annots = PdfArray()
        basepage.Annots += rmpage.Annots



from .generic import GenericPen
from .pencil import PencilPen
from .mechanicalpencil import MechanicalPencilPen
from .paintbrush import PaintbrushPen
from .ballpoint import BallpointPen
from .marker import MarkerPen
from .fineliner import FinelinerPen
from .highlighter import HighlighterPen
from .eraser import EraserPen
from .calligraphy import CalligraphyPen

# These pen codes probably refer to different versions through
# various system software updates. We'll just render them all
# the same (across all versions).
PEN_MAPPING = dict(enumerate([
    PaintbrushPen,       # Brush
    PencilPen,           # Pencil
    BallpointPen,        # Ballpoint
    MarkerPen,           # Marker
    FinelinerPen,        # Fineliner
    HighlighterPen,      # Highlighter
    EraserPen,           # Eraser
    MechanicalPencilPen, # Mechanical Pencil
    EraserPen,           # Erase Area
    None,                # unknown
    None,                # unknown
    None,                # unknown
    PaintbrushPen,       # Brush
    MechanicalPencilPen, # Mechanical Pencil
    PencilPen,           # Pencil
    BallpointPen,        # Ballpoint
    MarkerPen,           # Marker
    FinelinerPen,        # Fineliner
    HighlighterPen,      # Highlighter
    EraserPen,           # Eraser
    None,                # unknown
    CalligraphyPen       # Calligraphy
]))



'''
Copyright (C) 2020  Davis Remmel
Copyright 2021 Robert Schroll

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from .generic import GenericPen

class HighlighterPen(GenericPen):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.layer = kwargs.get('layer')
        self.annotate = False #TODO bool(int(QSettings().value(
        #    'pane/notebooks/export_pdf_annotate')))

    def paint_stroke(self, canvas, stroke):
        canvas.saveState()
        canvas.setLineCap(2)  # Square
        canvas.setLineJoin(1)  # Round
        #canvas.setDash ?? for solid line
        canvas.setStrokeColor((1.000, 0.914, 0.290), alpha=0.392)
        canvas.setLineWidth(stroke.width)

        path = canvas.beginPath()
        path.moveTo(stroke.segments[0].x, stroke.segments[0].y)
        for segment in stroke.segments[1:]:
            path.lineTo(segment.x, segment.y)
        canvas.drawPath(path, stroke=1, fill=0)
        canvas.restoreState()

        if self.annotate:
            assert False
            # Create outline of the path. Annotations that are close to
            # each other get groups. This is determined by overlapping
            # paths. In order to fuzz this, we'll double the normal
            # width and extend the end caps.
            self.setWidthF(self.widthF() * 2)
            self.setCapStyle(Qt.SquareCap)
            opath = QPainterPathStroker(self).createStroke(path)
            # The annotation type is carried all the way through. This
            # is the type specified in the PDF spec.
            self.layer.annot_paths.append(('Highlight', opath))



'''
Copyright (C) 2020  Davis Remmel
Copyright 2021 Robert Schroll

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

def pairs(iterable):
    it = iter(iterable)
    old = next(it)
    for new in it:
        yield (old, new)
        old = new

class GenericPen(object):
    def __init__(self, color, *args, **kwargs):
        self.color = color

    def paint_stroke(self, canvas, stroke):
        canvas.saveState()
        canvas.setLineCap(1)  # Rounded
        canvas.setLineJoin(1)  # Round join
        #canvas.setDash ?? for solid line
        canvas.setStrokeColor(self.color)
        for p1, p2 in pairs(stroke.segments):
            self.set_segment_properties(canvas, p1, p2)
            canvas.line(p1.x, p1.y, p2.x, p2.y)
        canvas.restoreState()

    def set_segment_properties(self, canvas, segment, nextsegment):
        # Set the width
        canvas.setLineWidth(segment.width)



'''
Copyright (C) 2020  Davis Remmel
Copyright 2021 Robert Schroll

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from .generic import GenericPen
from .textures import PENCIL_TEXTURES

class PencilPen(GenericPen):

    def set_segment_properties(self, canvas, segment, nextsegment):
        basewidth = segment.width
        deltamax = 0.42 * basewidth
        delta = -deltamax
        prim_width = basewidth + delta
        canvas.setLineWidth(prim_width)

        stroke_color = [1 - (1 - c) * segment.pressure for c in self.color]
        canvas.setStrokeColor(stroke_color)

    def old_paint_stroke(self, painter, stroke):
        assert False
        brush = QBrush()
        brush.setColor(self.color())

        # if self.vector:
        #     path = QPainterPath()
        #     path.moveTo(stroke.segments[0].x, stroke.segments[0].y)
        #     for i, segment in enumerate(stroke.segments, 1):
        #         path.lineTo(segment.x, segment.y)
        #     self.setWidthF(stroke.width)
        #     painter.setPen(self)
        #     painter.drawPath(path)
        #     return

        for i, segment in enumerate(stroke.segments):
            if i+1 >= len(stroke.segments):
                # no next segment, last 'to' point
                break

            nextsegment = stroke.segments[i+1]

            # # Direction has something to do with the brush shape...
            # print(segment.direction)

            # I experimented with a weighted pressure, but I don't think
            # the rM uses it! So, use a static pressure.
            # Experiment details: look at previous _n_ segment pressures
            # and average, where n={5,10,20,50,100,200) -- all this does
            # is blur the pressures, and all make it look further than
            # the truth.

            # # If the segment is short, use a round cap.
            # distance = point_distance(segment.x, segment.y,
            #                           nextsegment.x, nextsegment.y)
            # if distance < newwidth / 2:
            #     self.setCapStyle(Qt.RoundCap)
            # else:
            #     self.setCapStyle(Qt.FlatCap)

            # There is a spatter around the pencil. We are going to draw
            # two strokes: one for the primary, and another for the
            # spatter. The spatter is drawn first, so that it appears
            # behind in the vector versions.

            # Draw primary stroke
            basewidth = segment.width
            deltamax = 0.42 * basewidth
            delta = -deltamax
            prim_width = basewidth + delta
            self.setWidthF(prim_width)

            if self.vector:
                if not self.ocolor:
                    self.ocolor = self.color()
                ncolor = QColor()
                ncolor.setRedF(1 - ((1 - self.ocolor.redF()) * segment.pressure))
                ncolor.setGreenF(1 - ((1 - self.ocolor.greenF()) * segment.pressure))
                ncolor.setBlueF(1 - ((1 - self.ocolor.blueF()) * segment.pressure))
                self.setColor(ncolor)
            else:
                texture = PENCIL_TEXTURES.get_log(segment.pressure)
                brush.setTextureImage(texture)
                self.setBrush(brush)
            painter.setPen(self)
            painter.drawLine(QLineF(segment.x, segment.y,
                                    nextsegment.x, nextsegment.y))

            # Draw spatter stroke, but only if not vector because there
            # are compositing problems.
            if not self.vector:
                spat_width = prim_width * 1.25
                self.setWidthF(spat_width)

                texture = PENCIL_TEXTURES.get_log(segment.pressure * 0.7)
                brush.setTextureImage(texture)
                self.setBrush(brush)
                painter.setPen(self)
                painter.drawLine(QLineF(segment.x, segment.y,
                                        nextsegment.x, nextsegment.y))



'''
Copyright 2021 Robert Schroll

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import io
from pathlib import Path
import zipfile


# A Source should implement two methods:
#
# open(filename, mode)
#    Returns a file handled for the filename, opened in specified mode.
#
# exists(filename)
#    Returns a boolean indicating whether the file exists.
#
# In both cases, the filename may include the string `{ID}`, which indicates
# the Remarkable ID for that particular document (a UUID).  Thus, the caller
# of these methods does not have to know the ID of a document; the Source is
# responsible for filling that in appropriately.

class FSSource:

    def __init__(self, base_dir, doc_id):
        self.base_dir = Path(base_dir)
        self.doc_id = doc_id

    def format_name(self, name):
        return self.base_dir / name.format(ID=self.doc_id)

    def open(self, fn, mode='r'):
        return self.format_name(fn).open(mode)

    def exists(self, fn):
        return self.format_name(fn).exists()


class ZipSource:

    def __init__(self, zip_file, encoding='utf-8'):
        self.zip_file = zip_file
        self.encoding = encoding
        for fn in self.zip_file.namelist():
            if fn.endswith('.content'):
                self.doc_id = fn[:-8]
                break
        else:
            raise FileNotFoundError('Could not find .content file')

    def format_name(self, name):
        return name.format(ID=self.doc_id)

    def open(self, fn, mode='r'):
        f = self.zip_file.open(self.format_name(fn), mode.strip('b'))
        if mode.endswith('b'):
            return f
        return io.TextIOWrapper(f, encoding=self.encoding)

    def exists(self, fn):
        try:
            self.zip_file.getinfo(self.format_name(fn))
            return True
        except KeyError:
            return False


def get_source(source):
    # Pass through objects that implement the source API
    if hasattr(source, 'open') and hasattr(source, 'exists'):
        return source

    error = FileNotFoundError(f"Could not find a source file from {source!r}")
    try:
        source_p = Path(source)
    except TypeError:
        raise error from None

    # Check if it's a zip file
    if source_p.is_file():
        with source_p.open('rb') as f:
            if f.read(4) == b'PK\x03\x04':
                zf = zipfile.ZipFile(source_p)
                return ZipSource(zf)

    # If there's a .content file, assume it's unpacked
    if source_p.with_suffix('.content').is_file():
        return FSSource(source_p.parent, source_p.stem)

    raise error



'''
Copyright (C) 2020  Davis Remmel
Copyright 2021 Robert Schroll

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from .generic import GenericPen

class MarkerPen(GenericPen):
    def set_segment_properties(self, canvas, segment, nextsegment):
        # Set the width
        # maxdelta = segment.width / 0.5
        # delta = (segment.pressure - 1) * maxdelta
        # self.setWidthF(segment.width + delta)
        canvas.setLineWidth(segment.width * 0.7)



'''
Copyright (C) 2020  Davis Remmel
Copyright 2021 Robert Schroll

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import math

from .generic import GenericPen
from .textures import PENCIL_TEXTURES

def point_distance(x1, y1, x2, y2):
    dist = math.sqrt((x2 - x1)**2 + (y2 - y1)**2)
    return dist

class PaintbrushPen(GenericPen):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.vector = kwargs.get('vector', False)

    def set_segment_properties(self, canvas, segment, nextsegment):
        # Set the width
        modwidth = segment.width * 0.75
        maxdelta = modwidth * 0.75
        delta = (segment.pressure - 1) * maxdelta
        newwidth = modwidth + delta
        canvas.setLineWidth(newwidth)

        # # We want textures only in a mid-range, with the high and
        # # low ends going to solid patterns (clamping).
        # press_mod = segment.pressure + 0.4
        # if segment.pressure < 0.25:
        #     press_mod = 0
        # elif segment.pressure > 0.7:
        #     press_mod = 1

        press_mod = segment.pressure

        # There is also some effect of speed...really fast movements
        # produce really light strokes.
        press_mod *= 2 - (segment.speed / 75)

        if self.vector:
            stroke_color = [1 - (1 - c) * press_mod / 2 for c in self.color]
            canvas.setStrokeColor(stroke_color)
        else:
            assert False
            angle = math.degrees(nextsegment.direction) + 90
            transform = QTransform().rotate(angle)

            texture = PENCIL_TEXTURES.get_log_paintbrush(press_mod)
            brush.setTextureImage(texture)
            brush.setTransform(transform)
            self.setBrush(brush)

        # If the segment is short, use a round cap.
        distance = point_distance(segment.x, segment.y,
                                    nextsegment.x, nextsegment.y)
        if distance < newwidth / 1:
            canvas.setLineCap(1)  # Rounded
        else:
            canvas.setLineCap(0)  # Flat



'''
Copyright 2021 Robert Schroll

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import argparse
import io
import sys
import zipfile

from . import render
from .constants import VERSION
from .sources import ZipSource

def main():
    parser = argparse.ArgumentParser(description="Render a PDF file from a Remarkable document.")
    parser.add_argument('input', help="Filename of zip file, or root-level unpacked file of document.  Use '-' to read zip file from stdin.")
    parser.add_argument('output', nargs='?', default='', help="Filename where PDF file should be written.  Omit to write to stdout.")
    parser.add_argument('--version', action='version', version=VERSION)
    args = parser.parse_args()

    source = args.input
    if source == '-':
        # zipfile needs to seek, so we need to read this all in
        source = ZipSource(zipfile.ZipFile(io.BytesIO(sys.stdin.buffer.read())))
    if args.output:
        fout = open(args.output, 'wb')
    else:
        fout = sys.stdout.buffer

    stream = render(source)
    fout.write(stream.read())
    fout.close()
    return 0

if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rmrl', 'rmrl.pens']

package_data = \
{'': ['*'],
 'rmrl.pens': ['paintbrush_png/*',
               'paintbrush_textures_log/*',
               'pencil_lin_png/*',
               'pencil_log_png/*',
               'pencil_textures_linear/*',
               'pencil_textures_log/*']}

install_requires = \
['pdfrw>=0.4,<0.5',
 'reportlab>=3.5.60,<4.0.0',
 'svglib>=1.0.1,<2.0.0',
 'xdg>=5.0.1,<6.0.0']

setup_kwargs = {
    'name': 'rmrl',
    'version': '0.1.1',
    'description': 'Render reMarkable documents to PDF',
    'long_description': "rmrl: reMarkable Rendering Library\n===================================\nrmrl is a Python library for rendering reMarkable documents to PDF files.\nIt takes the original PDF document and the files describing your annotations,\ncombining them to produce a document close to what reMarkable itself would\noutput.\n\nDemo\n----\nThe same notebook was rendered to a PDF via the reMarkable app and rmrl.\nThe resultant PDF files were converted to PNGs with ImageMagick at 300\ndpi.\n\n reMarkable output | rmrl output\n:-----------------:|:-----------:\n[![reMarkable](demo/app.png)](demo/app.png) | [![rmrl](demo/rmrl.png)](demo/rmrl.png)\n\nThe biggest differences are the lack of texture in the pencils and paintbrush,\nwhich we hope to address in the future.  The highlight color is also different,\nbut we feel the default color is too subtle.\n\nInstallation\n------------\nrmrl requires Python 3.7 or later.  If that's installed, the easiest installation\nis to do a\n```bash\npip install rmrl\n```\nAlternatively, you may clone this repository.  [Poetry](https://python-poetry.org/) is used for development, so once that is installed you can run\n```bash\npoetry install\n```\nto get a virtual environment all set up.\n\nUsage\n-----\nThe main interface to rmrl is through a single function:\n```python\nfrom rmrl import render\n\noutput = render(source)\n```\n`source` may be:\n- The filename of a zip file containing the document.\n- The filename of any (root-level) file from an unpacked document.\n- Any object that provides `open()` and `exists()` methods.  See\n  `rmrl/sources.py` for more details on this API.\n\nThe output is a filestream with the contents of the PDF file.\n\nThe `render` function takes the following keyword arguments:\n- `progress_cb`: A callback function to be called periodically during the\n  rendering process.  It will be called with a single argument, a number\n  from 0 to 100 indicating the progress.  This function can abort the\n  process by raising an exception.\n\nCommand-line Usage\n------------------\nrmrl may be called as a command-line tool.  Once it has been installed, run\n```bash\npython -m rmrl filename\n```\nto convert `filename` to an annotated PDF.  The default output is to stdout.\nUse\n```bash\npython -m rmrl -h\n```\nto see all of the options.\n\nTemplates\n---------\nrmrl can use the reMarkable templates as a background when rendering notebooks.\nWe cannot ship copies of these templates.  You may be allowed to copy them from\nyour own reMarkable device on to your computer for personal use.  If this is\nlegal in your jurisdiction, you may connect your device to your computer by the\nUSB cable and run\n```bash\npython -m rmrl.load_templates\n```\nThis will copy these templates to `~/.local/share/rmrl/templates` (assuming\ndefault XDG settings).\n\nHistory\n-------\nrmrl derives from the [reMarkable Connection Utility](http://www.davisr.me/projects/rcu/),\nby Davis Remmel.  RCU is a full-featured GUI for managing all aspects of a\nreMarkable device.  Do check it out if you are looking for a stand-alone\nsolution for getting documents on and off of your device.\n\nRCU was chosen as a base for rmrl due to its high-quality rendering.  The\nfollowing are the major changes:\n- rmrl is designed as a library, for incorporation into other programs.  RCU\n  is designed as a stand-alone program.\n- rmrl uses the pure-Python [ReportLab Toolkit](https://www.reportlab.com/dev/opensource/rl-toolkit/)\n  for rendering PDF files.  RCU uses the Qt framework, which is a significantly\n  heavier installation.\n- rmrl only supports vector output, while RCU offers both raster and vector\n  rendering.\n- RCU supports PDF layers (Optional Content Groups).  At this point, rmrl does\n  not.\n- RCU can add PDF annotations corresponding to highlights.  At this point, rmrl\n  does not.\n\nTrademarks\n----------\nreMarkable(R) is a registered trademark of reMarkable AS. rmrl is not\naffiliated with, or endorsed by, reMarkable AS. The use of “reMarkable”\nin this work refers to the company’s e-paper tablet product(s).\n\nCopyright\n---------\nCopyright (C) 2020  Davis Remmel\n\nCopyright 2021 Robert Schroll\n\nThis program is free software: you can redistribute it and/or modify\nit under the terms of the GNU General Public License as published by\nthe Free Software Foundation, either version 3 of the License, or\n(at your option) any later version.\n\nThis program is distributed in the hope that it will be useful,\nbut WITHOUT ANY WARRANTY; without even the implied warranty of\nMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\nGNU General Public License for more details.\n\nYou should have received a copy of the GNU General Public License\nalong with this program.  If not, see <https://www.gnu.org/licenses/>.\n",
    'author': 'Robert Schroll',
    'author_email': 'rschroll@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rschroll/rmrl',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



'''
Copyright (C) 2020  Davis Remmel
Copyright 2021 Robert Schroll

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import gc
import json
import logging

from reportlab.graphics import renderPDF
from svglib.svglib import svg2rlg

from . import lines, pens
from .constants import DISPLAY, PDFHEIGHT, PDFWIDTH, PTPERPX, TEMPLATE_PATH


log = logging.getLogger(__name__)

class DocumentPage:
    # A single page in a document
    def __init__(self, source, pid, pagenum):
        # Page 0 is the first page!
        self.source = source
        self.num = pagenum

        # On disk, these files are named by a UUID
        self.rmpath = f'{{ID}}/{pid}.rm'
        if not source.exists(self.rmpath):
            # From the API, these files are just numbered
            pid = str(pagenum)
            self.rmpath = f'{{ID}}/{pid}.rm'

        # Try to load page metadata
        self.metadict = None
        metafilepath = f'{{ID}}/{pid}-metadata.json'
        if source.exists(metafilepath):
            with source.open(metafilepath, 'r') as f:
                self.metadict = json.load(f)

        # Try to load template
        self.template = None
        template_names = []
        pagedatapath = '{ID}.pagedata'
        if source.exists(pagedatapath):
            with source.open(pagedatapath, 'r') as f:
                template_names = f.read().splitlines()

        if template_names:
            # I have encountered an issue with some PDF files, where the
            # rM won't save the page template for later pages. In this
            # case, just take the last-available page template, which
            # is usually 'Blank'.
            template_name = template_names[max(self.num, len(template_names) - 1)]
            template_path = TEMPLATE_PATH / f'{template_name}.svg'
            if template_name != 'Blank' and template_path.exists():
                self.template = str(template_path)

        # Load layers
        self.layers = []
        self.load_layers()

    def get_grouped_annotations(self):
        # Return the annotations grouped by proximity. If they are
        # within a distance of each other, count them as a single
        # annotation.

        # Annotations should be delivered in an array, where each
        # index is a tuple (LayerName,
        annotations = []
        for layer in self.layers:
            annotations.append(layer.get_grouped_annotations())
        return annotations

    def load_layers(self):
        # Loads layers from the .rm files

        if not self.source.exists(self.rmpath):
            # no layers, obv
            return

        # Load reMy version of page layers
        pagelayers = None
        with self.source.open(self.rmpath, 'rb') as f:
            _, pagelayers = lines.readLines(f)

        # Load layer data
        for i in range(0, len(pagelayers)):
            layerstrokes = pagelayers[i]

            try:
                name = self.metadict['layers'][i]['name']
            except:
                name = 'Layer ' + str(i + 1)

            layer = DocumentPageLayer(self, name=name)
            layer.strokes = layerstrokes
            self.layers.append(layer)

    def render_to_painter(self, canvas, vector):
        # Render template layer
        if self.template:
            background = svg2rlg(self.template)
            background.scale(PDFWIDTH / background.width, PDFWIDTH / background.width)
            renderPDF.draw(background, canvas, 0, 0)
            # Bitmaps are rendered into the PDF as XObjects, which are
            # easy to pick out for layers. Vectors will render
            # everything inline, and so we need to add a 'magic point'
            # to mark the end of the template layer.
            if False and vector:  #TODO
                pen = GenericPen(color=Qt.transparent, vector=vector)
                painter.setPen(pen)
                painter.drawPoint(800, 85)

        # The annotation coordinate system is upside down compared to the PDF
        # coordinate system, so offset the bottom to the top and then flip
        # vertically along the old bottom / new top to place the annotations
        # correctly.
        canvas.translate(0, PDFHEIGHT)
        canvas.scale(PTPERPX, -PTPERPX)
        # Render user layers
        for layer in self.layers:
            # Bitmaps are rendered into the PDF as XObjects, which are
            # easy to pick out for layers. Vectors will render
            # everything inline, and so we need to add a 'magic point'
            # to mark the beginning of layers.
            if False and vector:  #TODO
                pen = GenericPen(color=Qt.transparent, vector=vector)
                painter.setPen(pen)
                painter.drawPoint(420, 69)
            layer.render_to_painter(canvas, vector)
        canvas.showPage()


class DocumentPageLayer:
    pen_widths = []

    def __init__(self, page, name=None):
        self.page = page
        self.name = name

        self.colors = [
            #QSettings().value('pane/notebooks/export_pdf_blackink'),
            #QSettings().value('pane/notebooks/export_pdf_grayink'),
            #QSettings().value('pane/notebooks/export_pdf_whiteink')
            (0, 0, 0),
            (0.5, 0.5, 0.5),
            (1, 1, 1)
        ]

        # Set this from the calling func
        self.strokes = None

        # Store PDF annotations with the layer, in case actual
        # PDF layers are ever implemented.
        self.annot_paths = []

    def get_grouped_annotations(self):
        # return: (LayerName, [(AnnotType, minX, minY, maxX, maxY)])

        # Compare all the annot_paths to each other. If any overlap,
        # they will be grouped together. This is done recursively.
        def grouping_func(pathset):
            newset = []

            for p in pathset:
                annotype = p[0]
                path = p[1]
                did_fit = False
                for i, g in enumerate(newset):
                    gannotype = g[0]
                    group = g[1]
                    # Only compare annotations of the same type
                    if gannotype != annotype:
                        continue
                    if path.intersects(group):
                        did_fit = True
                        newset[i] = (annotype, group.united(path))
                        break
                if did_fit:
                    continue
                # Didn't fit, so place into a new group
                newset.append(p)

            if len(newset) != len(pathset):
                # Might have stuff left to group
                return grouping_func(newset)
            else:
                # Nothing was grouped, so done
                return newset

        grouped = grouping_func(self.annot_paths)

        # Get the bounding rect of each group, which sets the PDF
        # annotation geometry.
        annot_rects = []
        for p in grouped:
            annotype = p[0]
            path = p[1]
            rect = path.boundingRect()
            annot = (annotype,
                     float(rect.x()),
                     float(rect.y()),
                     float(rect.x() + rect.width()),
                     float(rect.y() + rect.height()))
            annot_rects.append(annot)

        return (self.name, annot_rects)

    def paint_strokes(self, canvas, vector):
        for stroke in self.strokes:
            pen, color, unk1, width, unk2, segments = stroke

            penclass = pens.PEN_MAPPING.get(pen)
            if penclass is None:
                log.error("Unknown pen code %d" % pen)
                penclass = pens.GenericPen

            qpen = penclass(vector=vector,
                            layer=self,
                            color=self.colors[color])

            # Do the needful
            qpen.paint_stroke(canvas, stroke)

    def render_to_painter(self, painter, vector):
        if vector: # Turn this on with vector otherwise off to get hybrid
            self.paint_strokes(painter, vector=vector)
            return

        assert False

        # I was having problems with QImage corruption (garbage data)
        # and memory leaking on large notebooks. I fixed this by giving
        # the QImage a reference array to pre-allocate RAM, then reset
        # the reference count after I'm done with it, so that it gets
        # cleaned up by the python garbage collector.

        devpx = DISPLAY['screenwidth'] \
            * DISPLAY['screenheight']
        bytepp = 4  # ARGB32
        qimage = QImage(b'\0' * devpx * bytepp,
                        DISPLAY['screenwidth'],
                        DISPLAY['screenheight'],
                        QImage.Format_ARGB32)

        imgpainter = QPainter(qimage)
        imgpainter.setRenderHint(QPainter.Antialiasing)
        #imgpainter.setRenderHint(QPainter.LosslessImageRendering)
        self.paint_strokes(imgpainter, vector=vector)
        imgpainter.end()

        painter.drawImage(0, 0, qimage)

        del imgpainter
        del qimage
        gc.collect()



from pathlib import Path
import pkg_resources

from xdg import xdg_data_home

# From rcu.py, with comment
# Todo: this should be based on the specific RM model
DISPLAY = {
            'screenwidth': 1404,
            'screenheight': 1872,
            'realwidth': 1408,
            'dpi': 226
            }
# Qt docs say 1 pt is always 1/72 inch
# Multiply ptperpx by pixels to convert to PDF coords
PTPERPX = 72 / DISPLAY['dpi']
PDFHEIGHT = DISPLAY['screenheight'] * PTPERPX
PDFWIDTH = DISPLAY['screenwidth'] * PTPERPX

SPOOL_MAX = 10 * 1024 * 1024

# TODO: parameterize
TEMPLATE_PATH = xdg_data_home() / 'rmrl' / 'templates'

VERSION = pkg_resources.get_distribution('rmrl').version



'''
Copyright 2021 Robert Schroll

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import argparse
import subprocess
import sys
import textwrap

from .constants import TEMPLATE_PATH, VERSION

def main():
    parser = argparse.ArgumentParser(description="Load the templates from a Remarkable device for use with rmrl")
    parser.add_argument('ip', nargs='?', default='10.11.99.1', help="""
        IP address of Remarkable device.  Defaults to the value used when
        plugged in via USB.  Possible values can be found under Settings >
        Help > Copyrights and licenses, under the GPLv3 Compliance section.""")
    parser.add_argument('--version', action='version', version=VERSION)
    args = parser.parse_args()

    print(textwrap.dedent(f"""
        About to connect to your Remarkable device at {args.ip}.

        If this is the first time SSHing into your Remarkable device, you may
        see a warning that the authenticity of the host can't be established.
        This is expected; type 'yes' to continue.

        If you have not set up SSH keys on your Remarkable device, you will be
        prompted to enter a password.  This is NOT your lock screen passcode.
        It can be found under Settings > Help > Copyrights and licences, under
        the GPLv3 Compliance section.
    """).strip())
    print("")

    TEMPLATE_PATH.mkdir(parents=True, exist_ok=True)
    try:
        completed = subprocess.run(['scp', f'root@{args.ip}:/usr/share/remarkable/templates/*.svg', TEMPLATE_PATH],
            stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)
    except FileNotFoundError:
        print("Could not find the 'scp' program.")
        print("This should be installed as part of SSH.")
        return 1

    if completed.returncode == 0:
        print("")
        print(f"Templates copied to {TEMPLATE_PATH}")
    else:
        print("")
        print(f"Error: Got return code of {completed.returncode}")
        print("The cause may be indicated in the output above.")
    return completed.returncode

if __name__ == '__main__':
    sys.exit(main())



#from PySide2.QtGui import QImage
from pathlib import Path

class PencilTextures:
    def __init__(self):
        # Load textures
        self.textures_linear_pencil = []
        texpath = Path(__file__).parent / Path('pencil_textures_linear')
        texpaths = sorted(texpath.glob('*.ppm'))
        for p in texpaths:
            name = p.stem
            img = QImage()
            img.load(str(p))
            self.textures_linear_pencil.append(img)

        self.textures_log_pencil = []
        texpath = Path(__file__).parent / Path('pencil_textures_log')
        texpaths = sorted(texpath.glob('*.ppm'))
        for p in texpaths:
            name = p.stem
            img = QImage()
            img.load(str(p))
            self.textures_log_pencil.append(img)

        self.textures_log_paintbrush = []
        texpath = Path(__file__).parent / Path('paintbrush_textures_log')
        texpaths = sorted(texpath.glob('*.ppm'))
        for p in texpaths:
            name = p.stem
            img = QImage()
            img.load(str(p))
            self.textures_log_paintbrush.append(img)

    def get_linear(self, val):
        scale = len(self.textures_linear_pencil)
        i = int(val * scale)
        if i < 0:
            i = 0
        if i >= scale:
            i = scale - 1
        return self.textures_linear_pencil[i]

    def get_log(self, val):
        scale = len(self.textures_log_pencil)
        # These values were reached by trial-and-error.
        if val < 0:
            val = 0
        i = int(0.25 * (val * scale)**1.21)
        if i < 0:
            i = 0
        if i >= scale:
            i = scale - 1
        return self.textures_log_pencil[i]

    def get_log_paintbrush(self, val):
        scale = len(self.textures_log_paintbrush)
        if val < 0:
            val = 0
        i = int(0.25 * (val * scale)**1.21)
        if i < 0:
            i = 0
        if i >= scale:
            i = scale - 1
        return self.textures_log_paintbrush[i]

# Load pencil textures (shared for brushes, takes a lot of time
# because there are many)
PENCIL_TEXTURES = None #PencilTextures()



'''
lines.py
This is the model for Lines, which come from RM files.

This file was written originally for the reMy project and modified for
RCU.

reMy is a file manager for the reMarkable tablet.
Copyright (C) 2020  Emanuele D'Osualdo.

RCU is a synchronization tool for the reMarkable Tablet.
Copyright (C) 2020  Davis Remmel

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''


from collections import namedtuple
import struct
import json

Layer = namedtuple('Layer', ['strokes', 'name'])

Stroke = namedtuple(
    'Stroke',
    ['pen', 'color', 'unk1', 'width', 'unk2', 'segments']
)
Segment = namedtuple(
    'Segment',
    ['x', 'y', 'speed', 'direction', 'width', 'pressure']
)

HEADER_START = b'reMarkable .lines file, version='
S_HEADER_PAGE = struct.Struct('<{}ss10s'.format(len(HEADER_START)))
S_PAGE = struct.Struct('<BBH')    # TODO: might be 'I'
S_LAYER = struct.Struct('<I')
S_STROKE_V3 = struct.Struct('<IIIfI')
S_STROKE_V5 = struct.Struct('<IIIfII')
S_SEGMENT = struct.Struct('<ffffff')


class UnsupportedVersion(Exception):
    pass
class InvalidFormat(Exception):
    pass

def readStruct(fmt, source):
    buff = source.read(fmt.size)
    return fmt.unpack(buff)

def readStroke3(source):
    pen, color, unk1, width, n_segments = readStruct(S_STROKE_V3, source)
    return (pen, color, unk1, width, 0, n_segments)

def readStroke5(source):
    return readStruct(S_STROKE_V5, source)

# source is a filedescriptor from which we can .read(N)
def readLines(source):
    try:

        header, ver, *_ = readStruct(S_HEADER_PAGE, source)
        if not header.startswith(HEADER_START):
            raise InvalidFormat("Header is invalid")
        ver = int(ver)
        if ver == 3:
            readStroke = readStroke3
        elif ver == 5:
            readStroke = readStroke5
        else:
            raise UnsupportedVersion("Remy supports notebooks in the version 3 and 5 format only")
        n_layers, _, _ = readStruct(S_PAGE, source)
        layers = []
        for l in range(n_layers):
            n_strokes, = readStruct(S_LAYER, source)
            strokes = []
            for s in range(n_strokes):
                pen, color, unk1, width, unk2, n_segments = readStroke(source)
                segments = []
                for i in range(n_segments):
                    x, y, speed, direction, width, pressure = readStruct(S_SEGMENT, source)
                    segments.append(Segment(x, y, speed, direction, width, pressure))
                strokes.append(Stroke(pen, color, unk1, width, unk2, segments))
            layers.append(strokes)

        return (ver, layers)

    except struct.error:
        raise InvalidFormat("Error while reading page")
